package aql

import (
	"fmt"
	"strconv"

	"git.science.uu.nl/graphpolaris/query-conversion/entity"
)

type pdict struct {
	typename string
	pointer  int
}

type pdictList []pdict

var listoflists []pdictList
var reldone map[int]bool
var entdone map[int]bool
var funcdone map[int]bool
var relfuncdone map[int]bool
var filterDone map[int]bool

func search(JSONQuery *entity.IncomingQueryJSON, index int) {
	listoflists = []pdictList{}
	reldone = make(map[int]bool)
	entdone = make(map[int]bool)
	funcdone = make(map[int]bool)
	relfuncdone = make(map[int]bool)
	filterDone = make(map[int]bool)
	var s pdictList
	//printSlice(s)
	//layercounter = 0

	initent := pdict{
		typename: "entity",
		pointer:  index,
	}

	s = append(s, initent)
	listoflists = append(listoflists, s)
	entToRel(JSONQuery, initent)

	for i := range listoflists {
		for j := range listoflists[i] {
			fmt.Println(listoflists[i][j])
		}
		fmt.Println("")
	}
	addFilters(JSONQuery)
	fmt.Println(listoflists)

}

/*
relToEnt Get the entities connected to a relation and recursivly constructs part of the hierarchy
Entities always get added IN FRONT OF their respective relation in the hierarchy
	JSONQuery: *entity.IncomingQueryJSON, the query in JSON format
	rel: pdict, the relation to find all connected entities for
*/
func relToEnt(JSONQuery *entity.IncomingQueryJSON, rel pdict) {
	var newlist pdictList
	layercounter := findCurrentLayer(listoflists, rel)
	// Loop over all entities
	// If an entity is already in the entdone dict we already added it to the hierarchy, so we don't have to add it again
	// If an entity matches either the from or to in a relation we can add it to the newlist
	for i := range JSONQuery.Entities {
		if _, ok := entdone[i]; !ok {
			if JSONQuery.Relations[rel.pointer].FromID == i && JSONQuery.Relations[rel.pointer].FromType == "entity" {
				fromentity := pdict{
					typename: "entity",
					pointer:  i,
				}
				newlist = append(newlist, fromentity)
			} else if JSONQuery.Relations[rel.pointer].ToID == i && JSONQuery.Relations[rel.pointer].ToType == "entity" {
				toentity := pdict{
					typename: "entity",
					pointer:  i,
				}
				newlist = append(newlist, toentity)
			}
		}
	}
	// This relation has found all its entities so we can set it's ID to true
	reldone[rel.pointer] = true
	// If the newlist is empty, we can just skip the recursion
	// This is effectively our base case
	if len(newlist) != 0 {
		// If our layercounter is equal to 0 we are in the first "layer" of the hierarchy
		// Because we add the entities IN FRONT OF their respective relation we don't have to move the layercounter before prepending
		// If our layercounter is not equal to 0 we lower the layercounter and then add each item to the newly selected layer
		listoflists = aboveAppend(listoflists, layercounter, newlist)

		// After getting a list of entities we can only go towards a list of relation
		// So we recurse by calling EntToRel
		for i := range newlist {
			fmt.Println("EntToRel being called with index?: " + strconv.Itoa(newlist[i].pointer))
			entToRel(JSONQuery, newlist[i])

		}
	}

}

/*
entToRel Get the relations connected to a entity and recursivly constructs part of the hierarchy
Relation always get added BEHIND their respective entity in the hierarchy
	JSONQuery: *entity.IncomingQueryJSON, the query in JSON format
	ent: pdict, the entity to find all connected relations for
*/
func entToRel(JSONQuery *entity.IncomingQueryJSON, ent pdict) {
	var newlist pdictList
	layercounter := findCurrentLayer(listoflists, ent)
	// Loop over all relations
	// If a relation is already in the reldone dict we already added it to the hierarchy, so we don't have to add it again
	// If a relation matches either the from or to with the entity we can add it to the newlist
	for i := range JSONQuery.Relations {
		if _, ok := reldone[i]; !ok {
			if JSONQuery.Relations[i].FromID == ent.pointer && JSONQuery.Relations[i].FromType == "entity" {
				rel := pdict{
					typename: "relation",
					pointer:  i,
				}
				newlist = append(newlist, rel)
			} else if JSONQuery.Relations[i].ToID == ent.pointer && JSONQuery.Relations[i].ToType == "entity" {
				rel := pdict{
					typename: "relation",
					pointer:  i,
				}
				newlist = append(newlist, rel)
			}
		}
	}
	// This entity has found all its relations so we can set it's ID to true
	entdone[ent.pointer] = true

	if len(newlist) != 0 {
		// If our layercounter is equal to the length of the hierarchy - 1 we are in the last "layer" of the hierarchy
		// Because we add the relations BEHIND their respective entities we don't have to move the layercounter before appending
		// TODO TAKE OUT UNNEEDED LAYERCOUNTER INCREMENTS AND DECREMENTS
		// If our layercounter is any other value we increase the layercounter and then add each item to the newly selected layer
		listoflists = belowAppend(listoflists, layercounter, newlist)

		// After getting a list of relations we can only go towards a list of entities or a list of functions
		// So we recurse by calling RelToEnt and RelToAllFunc
		for i := range newlist {
			fmt.Println("RelToEnt being called with index?: " + strconv.Itoa(newlist[i].pointer))
			relToEnt(JSONQuery, newlist[i])
			fmt.Println("RelToAllFunc being called with index?: " + strconv.Itoa(newlist[i].pointer))
			relToAllFunc(JSONQuery, newlist[i])

		}
	}
}

/*
relToAllFunc Get the functions connected (both functions that are applied to a subquery a relation is a part of and functions the relation is connected to itself)
 to a relation and recursivly constructs part of the hierarchy
If a function is applied to a subquery the relation is a part of, we add it BEHIND its respective relation
If a function is connected to a relation (relation uses the results from the function), we add it IN FRONT OF its respective relation
	JSONQuery: *entity.IncomingQueryJSON, the query in JSON format
	rel: pdict, the relation to find all connected functions for
*/
func relToAllFunc(JSONQuery *entity.IncomingQueryJSON, rel pdict) {
	var funcappliedtosubquery pdictList
	var functowhichrelapplies pdictList
	layercounter := findCurrentLayer(listoflists, rel)
	// Loop over all functions
	// If a relation is already in the relfuncdone dict we already added it to the hierarchy, so we don't have to add it again
	// If a function's relationID matches the current relation then the function is applied to a subquery
	// If the relation's functionpointer matches a function's ID then the relation is connected to the function
	// Depending on the case they get put in a different list and are put in different places in the hierarchy
	for i := range JSONQuery.GroupBys {
		if _, ok := relfuncdone[rel.pointer]; !ok {
			if _, ok := funcdone[i]; !ok {
				if JSONQuery.GroupBys[i].RelationID == rel.pointer {
					relfunc := pdict{
						typename: "groupBy",
						pointer:  i,
					}
					funcappliedtosubquery = append(funcappliedtosubquery, relfunc)
					fmt.Println("I AM HERE 1")
				}

				if JSONQuery.Relations[rel.pointer].FromID == i && JSONQuery.Relations[rel.pointer].FromType == "groupBy" {
					fromfunc := pdict{
						typename: "groupBy",
						pointer:  i,
					}
					functowhichrelapplies = append(functowhichrelapplies, fromfunc)
					fmt.Println("I AM HERE 2")

				} else if JSONQuery.Relations[rel.pointer].ToID == i && JSONQuery.Relations[rel.pointer].ToType == "groupBy" {
					tofunc := pdict{
						typename: "groupBy",
						pointer:  i,
					}
					functowhichrelapplies = append(functowhichrelapplies, tofunc)
					fmt.Println("I AM HERE 3")

				}

			}
		}

	}
	relfuncdone[rel.pointer] = true
	layercountertwo := layercounter
	layercounterthree := layercounter
	// See main function comment to see which sublist gets put where in the hierarchy
	if len(functowhichrelapplies) != 0 {
		listoflists = aboveAppend(listoflists, layercountertwo, functowhichrelapplies)

		for i := range functowhichrelapplies {
			fmt.Println("FuncToAllRell being called with index?: " + strconv.Itoa(functowhichrelapplies[i].pointer))
			funcToAllRel(JSONQuery, functowhichrelapplies[i])

		}
	}

	if len(funcappliedtosubquery) != 0 {
		//newlayercounter := layercounter
		listoflists = belowAppend(listoflists, layercounterthree, funcappliedtosubquery)

		for i := range funcappliedtosubquery {
			fmt.Println("FuncToAllRel being called with index?: " + strconv.Itoa(funcappliedtosubquery[i].pointer))
			funcToAllRel(JSONQuery, funcappliedtosubquery[i])
		}
	}

}

/*
funcToAllRel Get the relations connected (both relations that are in a subquery a function is applied to and relations that are connected to the function itself)
 to a function and recursivly constructs part of the hierarchy
If a relation is in a subquery that the function is applied to, we add the relation IN FRONT OF its respective function
If a relation is connected to a function, we add the relation BEHIND its respective function
	JSONQuery: *entity.IncomingQueryJSON, the query in JSON format
	function: pdict, the function to find all connected relations for
*/
func funcToAllRel(JSONQuery *entity.IncomingQueryJSON, function pdict) {
	var funcappliedtosubquery pdictList
	var relattachedtofunc pdictList
	layercounter := findCurrentLayer(listoflists, function)
	for i := range JSONQuery.Relations {
		if _, ok := funcdone[function.pointer]; !ok {
			if _, ok := relfuncdone[i]; !ok {
				// The func is attached to this relation
				if JSONQuery.GroupBys[function.pointer].RelationID == i {
					funcrel := pdict{
						typename: "relation",
						pointer:  i,
					}
					funcappliedtosubquery = append(funcappliedtosubquery, funcrel)

				}

				if JSONQuery.Relations[i].FromID == function.pointer && JSONQuery.Relations[i].FromType == "groupBy" {
					fromrel := pdict{
						typename: "relation",
						pointer:  i,
					}
					relattachedtofunc = append(relattachedtofunc, fromrel)

				} else if JSONQuery.Relations[i].ToID == function.pointer && JSONQuery.Relations[i].ToType == "groupBy" {
					torel := pdict{
						typename: "relation",
						pointer:  i,
					}
					relattachedtofunc = append(relattachedtofunc, torel)

				}

			}
		}
	}
	funcdone[function.pointer] = true

	layercountertwo := layercounter
	layercounterthree := layercounter
	if len(funcappliedtosubquery) != 0 {
		//newlayercounter := layercounter
		listoflists = aboveAppend(listoflists, layercountertwo, funcappliedtosubquery)

		for i := range funcappliedtosubquery {
			fmt.Println("RelToEnt being called with index?: " + strconv.Itoa(funcappliedtosubquery[i].pointer))
			relToEnt(JSONQuery, funcappliedtosubquery[i])
			fmt.Println("RelToAllFunc being called with index?: " + strconv.Itoa(funcappliedtosubquery[i].pointer))
			relToAllFunc(JSONQuery, funcappliedtosubquery[i])

		}
	}

	if len(relattachedtofunc) != 0 {
		listoflists = belowAppend(listoflists, layercounterthree, relattachedtofunc)

		for i := range relattachedtofunc {
			fmt.Println("RelToEnt being called with index?: " + strconv.Itoa(relattachedtofunc[i].pointer))
			relToEnt(JSONQuery, relattachedtofunc[i])
			fmt.Println("RelToAllFunc being called with index?: " + strconv.Itoa(relattachedtofunc[i].pointer))
			relToAllFunc(JSONQuery, relattachedtofunc[i])
		}
	}
}

func addFilters(JSONQuery *entity.IncomingQueryJSON) {

	for i, filter := range JSONQuery.Filters {
		if _, ok := filterDone[i]; !ok {
			p := pdict{
				typename: filter.FromType,
				pointer:  filter.FromID,
			}
			f := pdict{
				typename: "filter",
				pointer:  filter.ID,
			}
			addOneFilter(f, JSONQuery, p, &filterDone)
		}
	}
}

func addOneFilter(filterPDict pdict, JSONQuery *entity.IncomingQueryJSON, p pdict, filterDone *map[int]bool) {
	if p.typename == "filter" && (*filterDone)[p.pointer] {
		l := findCurrentLayer(listoflists, p)
		k := pdictList{filterPDict}
		if len(listoflists) > l+1 && listoflists[l+1][0].typename == "filter" {
			listoflists[l+1] = append(listoflists[l+1], filterPDict)
		} else {
			listoflists = filterAppend(listoflists, l, k)
		}
		(*filterDone)[filterPDict.pointer] = true
	} else if p.typename == "filter" {
		pnew := pdict{
			typename: JSONQuery.Filters[p.pointer].FromType,
			pointer:  JSONQuery.Filters[p.pointer].FromID,
		}
		addOneFilter(p, JSONQuery, pnew, filterDone)
		l := findCurrentLayer(listoflists, p)
		k := pdictList{filterPDict}
		if len(listoflists) > l+1 && listoflists[l+1][0].typename == "filter" {
			listoflists[l+1] = append(listoflists[l+1], filterPDict)
		} else {
			listoflists = filterAppend(listoflists, l, k)
		}
		(*filterDone)[filterPDict.pointer] = true
	} else {
		l := findCurrentLayer(listoflists, p)
		k := pdictList{filterPDict}
		if len(listoflists) > l+1 && listoflists[l+1][0].typename == "filter" {
			listoflists[l+1] = append(listoflists[l+1], filterPDict)
		} else {
			listoflists = filterAppend(listoflists, l, k)
		}
		(*filterDone)[filterPDict.pointer] = true
	}
}

// A function that appends 1 level above (if index is 0 this won't work)
func aboveAppend(lists []pdictList, index int, values pdictList) []pdictList {
	if index == 0 {
		return prepend(lists, values)
	} else {
		index--
		for i := range values {
			lists[index] = append(lists[index], values[i])
		}
		return lists
	}

}

// A function that appends 1 level below
func belowAppend(lists []pdictList, index int, values pdictList) []pdictList {
	if index == len(listoflists)-1 {
		lists = append(listoflists, values)
		return lists

	} else {
		index++
		for i := range values {
			lists[index] = append(lists[index], values[i])
		}
		return lists

	}
}

func filterAppend(lists []pdictList, index int, values pdictList) []pdictList {
	if len(lists)-1 == index { // nil or empty slice or after last element
		return append(lists, values)
	}
	k := make([]pdictList, len(lists[index+1:]))
	copy(k, lists[index+1:])
	l := make([]pdictList, len(lists[:index+1]))
	copy(l, lists[:index+1])
	lists = append(l, values) // index < len(a)
	return append(lists, k...)

}

// A simple double-for loop that finds the layer in which an element resides in the hierarchy
// Because we only append elements relative to another element, we can freely use this to keep track of layers
func findCurrentLayer(list []pdictList, element pdict) int {
	currlayer := -1
	for i, sublist := range list {
		for j := range sublist {
			if sublist[j].pointer == element.pointer && sublist[j].typename == element.typename {
				currlayer = i
				//break
			}
		}
	}
	return currlayer
}

// Adds a list of pdicts to the hierarchy, but IN FRONT OF the current layer
// Only needed when a entire new list has to be inserted in front of the hierarcy
// Prepending to existing layers can be done by decreasing the layercounter and appending
// See XToY functions for example usage
func prepend(list []pdictList, element pdictList) []pdictList {
	var dummylist pdictList
	dummy := pdict{
		typename: "dummy",
		pointer:  -1,
	}
	dummylist = append(dummylist, dummy)
	list = append(list, dummylist)
	copy(list[1:], list)
	list[0] = element
	return list
}
